function UKFML()
%% UKFML
% Adapted from code by Franz Hamilton, Tyrus Berry, and Tim Sauer (George Mason University).
% Produces example figure for initializing with HOPF regime parameters for
% estimating parameters associated with SNIC regime.
%%% Preliminary setup
obsNoise=.01; s=4; rng(s);
T = 200000; dt=.1; dT=.1; nn=fix(dT/dt);
%%% Define parameters
truep = [.067,4,12,17.4,120,-84,-60,8,2,-1.2,18,20,100];
guessp= [.04,4,2,30,120,-84,-60,8,2,-1.2,18,20,100];
SimName='tSNICgHOPF';
EstSet={'V','n','phi','gCa','V3','V4','gK','gL','V1','V2'};
dx = length(EstSet);
icovmat=.001*eye(dx); % prior, P_{{xx}_0}
lambda=5;
%%% Generate Data
x00=[-4 .14]';
parameters=ones(T+1,1)*truep; time=0:dT:dT*T;
x0=generateMLME(x00,T,parameters,dT); x0=generateMLME(x0(:,end),T,parameters,dT);
signalStdev=std(x0(1,:));
ML_Noisy=x0(1,:)+obsNoise*(signalStdev.*randn(size(x0(1,:))));
obs = ML_Noisy';
%%% Define initial guesses for state components
istate = zeros(dx,1)*10;
istate(2)=0; istate(1)=obs(1); istate(3:end)=[guessp(1:4) guessp(8:11)];
%%% Define best guesses for Q,R
Vrange=max(obs)-min(obs);
Qscale=1e-7; Q = Qscale*diag([Vrange, 1, abs(istate(3:end))']);
R = (obsNoise*signalStdev)^2; if R==0, R=1e-4; end
redistEnsemble=1; % Can make 0 and compare performance
%%% Run Filter
[stateEstimate,variance]= ...
    UKF_Estimate(time,istate,obs',@MLstatefctModEuler,@MLstateobsfct,Q,R,...
    parameters(1:end,:),lambda,icovmat,dt,nn,redistEnsemble);
%%% Make parameter estimation plot
for i =3:dx
    figure(11)
    plot(time,stateEstimate(i,:),'linewidth',3)
    hold on
end
paramLegend={'\phi' ,   'g_{Ca}' ,   'V_3'  ,  'V_4'  ,  'g_{K}' ,...
    'g_l' ,   'V_1' ,'V_2'};
xlim([0 time(end)])
legend(paramLegend)
set(gca,'FontSize',16)
xlabel('Time (ms)','FontSize',25)
end
function [fstate,fcovmat,innovation] = UKF_STEP(itime,ftime,istate,...
    icovmat,fobs,F,H,Q,R,p,lambda,t,dt,nn,redistEnsemble)
%%%Propagate from one observation time step to the next
% itime   :     the current time
% ftime   :     the final time for this step of the filter
% istate  :     an Nx1 vector containing the state at itime
% icovmat :     an NxN matrix containing the covariance matrix at itime
% fobs    :     an Mx1 vector which contains the observation at ftime
% F       :     a function handle which computes the vector field as a
%                function of the state and time
% H       :     a function handle which computes the observation as a
%                function of the state and time
% Q       :     the NxN covariance matrix for the dynamical noise
% R       :      the MxM covariance matrix for the observation noise
% p       :     fixed parameter set for the model. initial guesses for
%                parameters to be estimated are included in istate.
% lambda  :     algorithmic parameter to spread or shrink ensemble
%
N = length(istate);
M = length(fobs);
%%% Fix up the covariance matrices
icovmat = (icovmat+icovmat')/2;
[u,l]=svd(Q); Q = u*l*u';
[u,l]=svd(R); R = u*l*u';
%%% Initialize Filter
xhat_a = istate;
Pxx = icovmat;
%%% Build the matrix square root and ensemble
[U,S,V] = svd(Pxx);
squareRoot = U*diag(sqrt(diag(S)))*U';
xEnsemble = zeros(N,2*N+1);
xEnsemble(:,1) = xhat_a;
xEnsemble(:,2:N+1) = repmat(xhat_a,1,N)+sqrt(N+lambda)*squareRoot;
xEnsemble(:,N+2:2*N+1) = repmat(xhat_a,1,N)-sqrt(N+lambda)*squareRoot;
W = [lambda/(N+lambda) 1/(2*(N+lambda))*ones(1,2*N)]; %weights
xUpdatedEnsemble = F(xEnsemble,p,t,dt,nn); % forward mapping
yEnsemble = H(xUpdatedEnsemble); % observational mapping
yhat_b = sum(repmat(W,M,1).*yEnsemble,2);
observationDifference = yEnsemble-repmat(yhat_b,1,2*N+1);
Pb = (W.*observationDifference)*observationDifference';

%%% Calculate the background state estimate
xhat_b = sum(repmat(W,N,1).*xUpdatedEnsemble,2);
%%% Calculate the background covariance estimate
ensembleDifference = xUpdatedEnsemble-repmat(xhat_b,1,2*N+1);
FPa = (W.*ensembleDifference)*ensembleDifference';
Pxx_f = FPa+Q;
Pxx_f = min(Pxx_f,2); %Restrict uncertainty, can remove

if redistEnsemble
    %%% Redistribute sigma points around background mean
    %%% Build the matrix square root and ensemble
    [U,S,V] = svd(Pxx_f);
    squareRoot = U*diag(sqrt(diag(S)))*U';
    xUpdatedEnsemble = zeros(N,2*N+1);
    xUpdatedEnsemble(:,1) = xhat_b;
    xUpdatedEnsemble(:,2:N+1) = repmat(xhat_b,1,N)+sqrt(N+lambda)*squareRoot;
    xUpdatedEnsemble(:,N+2:2*N+1) = repmat(xhat_b,1,N)-sqrt(N+lambda)*squareRoot;
    %%% Realculate the background state estimate
    xhat_b = sum(repmat(W,N,1).*xUpdatedEnsemble,2);
    %%% Recalculate the background covariance estimate
    ensembleDifference = xUpdatedEnsemble-repmat(xhat_b,1,2*N+1);
    Pxx_f = (W.*ensembleDifference)*ensembleDifference';
    %%% Calculate predicted measurements
    yEnsemble = H(xUpdatedEnsemble);
    yhat_b = sum(repmat(W,M,1).*yEnsemble,2);
observationDifference = yEnsemble-repmat(yhat_b,1,2*N+1);
Pb = ((W.*observationDifference)*observationDifference');
end
%%% Construct the Kalman gain
Pyy = Pb+R;
Pxy = (W.*ensembleDifference)*observationDifference';
K = Pxy/Pyy;
innovation = (fobs-yhat_b);
%%% Analysis (estimation) step
xhat_a = xhat_b+K*innovation;
Pxx = Pxx_f - K*Pyy*K'; % same as Pxx_f -KPxy'
fcovmat = Pxx;
fstate = xhat_a;
end

function r=MLstatefctModEuler(x,p,j,dt,nn)
xnl=x;
for n=1:nn
    k1=dt*fXaug(0,xnl,p(j,:));
    k2=dt*fXaug(0,xnl+k1,p(j+1,:));
    xnl=xnl+k1./2+k2./2;
end
r=xnl;
end

function Hx=MLstateobsfct(x)
Hx=x(1,:);
end
function [stateEstimate,varEstimate] = UKF_Estimate(time,istate,obs,F,H,...
    Qguess,Rguess,p,lambda,icovmat,dt,nn,redistEnsemble)
N = length(istate); M = size(obs,1); T = size(obs,2);
Q = Qguess; R = Rguess;
stateEstimate = zeros(N,T); stateEstimate(:,1)=istate;
varEstimate = zeros(N,T); varEstimate(:,1)=diag(icovmat);
innovseq = zeros(M,T);
for t = 1:T-1
    %%% One step of UKF (t_k -> t_{k+1})
    [istate,icovmat,innovation] = UKF_STEP(time(t),time(t+1),istate,...
        icovmat,obs(:,t+1),F,H,Q,R,p,lambda,t,dt,nn,redistEnsemble);
    stateEstimate(:,t+1) = istate;
    varEstimate(:,t+1) = diag(icovmat);
    innovseq(:,t+1) = innovation;
end
end